#pragma once

#include "async.h"
#include "common.h"
#include "io_helpers.h"

#include <string>
#include <variant>

namespace saw {
/*
 * Input stream
 */
class InputStream {
public:
	virtual ~InputStream() = default;

	virtual error_or<size_t> read(void *buffer, size_t length) = 0;

	virtual conveyor<void> readReady() = 0;

	virtual conveyor<void> onReadDisconnected() = 0;
};

/*
 * Output stream
 */
class OutputStream {
public:
	virtual ~OutputStream() = default;

	virtual error_or<size_t> write(const void *buffer, size_t length) = 0;

	virtual conveyor<void> writeReady() = 0;
};

/*
 * Io stream
 */
class IoStream : public InputStream, public OutputStream {
public:
	virtual ~IoStream() = default;
};

class AsyncInputStream {
public:
	virtual ~AsyncInputStream() = default;

	virtual void read(void *buffer, size_t min_length, size_t max_length) = 0;

	virtual conveyor<size_t> readDone() = 0;
	virtual conveyor<void> onReadDisconnected() = 0;
};

class AsyncOutputStream {
public:
	virtual ~AsyncOutputStream() = default;

	virtual void write(const void *buffer, size_t length) = 0;

	virtual conveyor<size_t> writeDone() = 0;
};

class AsyncIoStream final : public AsyncInputStream, public AsyncOutputStream {
private:
	own<IoStream> stream;

	conveyor_sink read_ready;
	conveyor_sink write_ready;
	conveyor_sink read_disconnected;

	ReadTaskAndStepHelper read_stepper;
	WriteTaskAndStepHelper write_stepper;

public:
	AsyncIoStream(own<IoStream> str);

	SAW_FORBID_COPY(AsyncIoStream);
	SAW_FORBID_MOVE(AsyncIoStream);

	void read(void *buffer, size_t length, size_t max_length) override;

	conveyor<size_t> readDone() override;

	conveyor<void> onReadDisconnected() override;

	void write(const void *buffer, size_t length) override;

	conveyor<size_t> writeDone() override;
};

class Server {
public:
	virtual ~Server() = default;

	virtual conveyor<own<IoStream>> accept() = 0;
};

class NetworkAddress;
/**
 * Datagram class. Bound to a local address it is able to receive inbound
 * datagram messages and send them as well as long as an address is provided as
 * well
 */
class Datagram {
public:
	virtual ~Datagram() = default;

	virtual error_or<size_t> read(void *buffer, size_t length) = 0;
	virtual conveyor<void> readReady() = 0;

	virtual error_or<size_t> write(const void *buffer, size_t length,
								   NetworkAddress &dest) = 0;
	virtual conveyor<void> writeReady() = 0;
};

class OsNetworkAddress;
class StringNetworkAddress;

class NetworkAddress {
public:
	using ChildVariant =
		std::variant<OsNetworkAddress *, StringNetworkAddress *>;

	virtual ~NetworkAddress() = default;

	virtual NetworkAddress::ChildVariant representation() = 0;

	virtual const std::string &address() const = 0;
	virtual uint16_t port() const = 0;
};

class OsNetworkAddress : public NetworkAddress {
public:
	virtual ~OsNetworkAddress() = default;

	NetworkAddress::ChildVariant representation() override { return this; }
};

class StringNetworkAddress final : public NetworkAddress {
private:
	std::string address_value;
	uint16_t port_value;

public:
	StringNetworkAddress(const std::string &address, uint16_t port);

	const std::string &address() const override;
	uint16_t port() const override;

	NetworkAddress::ChildVariant representation() override { return this; }
};

class Network {
public:
	virtual ~Network() = default;

	/**
	 * Resolve the provided string and uint16 to the preferred storage method
	 */
	virtual conveyor<own<NetworkAddress>>
	resolveAddress(const std::string &addr, uint16_t port_hint = 0) = 0;

	/**
	 * Parse the provided string and uint16 to the preferred storage method
	 * Since no dns request is made here, no async conveyors have to be used.
	 */
	/// @todo implement
	// virtual own<NetworkAddress> parseAddress(const std::string& addr,
	// uint16_t port_hint = 0) = 0;

	/**
	 * Set up a listener on this address
	 */
	virtual own<Server> listen(NetworkAddress &bind_addr) = 0;

	/**
	 * Connect to a remote address
	 */
	virtual conveyor<own<IoStream>> connect(NetworkAddress &address) = 0;

	/**
	 * Bind a datagram socket at this address.
	 */
	virtual own<Datagram> datagram(NetworkAddress &address) = 0;
};

class IoProvider {
public:
	virtual ~IoProvider() = default;

	virtual own<InputStream> wrapInputFd(int fd) = 0;

	virtual Network &network() = 0;
};

struct AsyncIoContext {
	own<IoProvider> io;
	event_loop &event_loop;
	event_port &event_port;
};

error_or<AsyncIoContext> setupAsyncIo();
} // namespace saw
