#pragma once

#include "buffer.h"
#include "message.h"
#include "stream_endian.h"

namespace saw {
class ProtoKelCodec {
public:
	using UnionIdT = uint32_t;
	using ArrayLengthT = uint64_t;
	using PacketLengthT = uint64_t;

private:
	struct ReadContext {
		Buffer &buffer;
		size_t offset = 0;
	};

	template <typename T> friend struct ProtoKelEncodeImpl;

	template <typename T> friend struct ProtoKelDecodeImpl;

public:
	struct Limits {
		ProtoKelCodec::PacketLengthT packet_size;

		Limits() : packet_size{4096} {}
		Limits(ProtoKelCodec::PacketLengthT ps) : packet_size{ps} {}
	};

	struct Version {
		size_t major;
		size_t minor;
		size_t security;
	};

	const Version version() const { return Version{0, 0, 0}; }

	template <class Schema, class Container = MessageContainer<Schema>>
	error encode(typename Message<Schema, Container>::Reader reader,
				 Buffer &buffer);

	template <class Schema, class Container = MessageContainer<Schema>>
	error decode(typename Message<Schema, Container>::Builder builder,
				 Buffer &buffer, const Limits &limits = Limits{});
};

template <class T> struct ProtoKelEncodeImpl;

template <class T, size_t N, class Container>
struct ProtoKelEncodeImpl<Message<schema::Primitive<T, N>, Container>> {
	static error
	encode(typename Message<schema::Primitive<T, N>, Container>::Reader data,
		   Buffer &buffer) {
		error err = StreamValue<typename PrimitiveTypeHelper<
			schema::Primitive<T, N>>::Type>::encode(data.get(), buffer);
		return err;
	}

	static size_t
	size(typename Message<schema::Primitive<T, N>, Container>::Reader) {
		return StreamValue<typename PrimitiveTypeHelper<
			schema::Primitive<T, N>>::Type>::size();
	}
};

template <class Container>
struct ProtoKelEncodeImpl<Message<schema::String, Container>> {
	static error
	encode(typename Message<schema::String, Container>::Reader data,
		   Buffer &buffer) {
		std::string_view view = data.get();
		size_t size = view.size();

		error err = buffer.writeRequireLength(sizeof(size) + size);
		if (err.failed()) {
			return err;
		}

		err = StreamValue<size_t>::encode(size, buffer);
		if (err.failed()) {
			return err;
		}
		for (size_t i = 0; i < view.size(); ++i) {
			buffer.write(i) = view[i];
		}
		buffer.writeAdvance(view.size());
		return no_error();
	}

	static size_t
	size(typename Message<schema::String, Container>::Reader reader) {
		return sizeof(size_t) + reader.get().size();
	}
};

template <class... T, class Container>
struct ProtoKelEncodeImpl<Message<schema::Tuple<T...>, Container>> {
	template <size_t i = 0>
	static typename std::enable_if<i == sizeof...(T), error>::type
	encodeMembers(typename Message<schema::Tuple<T...>, Container>::Reader,
				  Buffer &) {
		return no_error();
	}

	template <size_t i = 0>
	static typename std::enable_if<(i < sizeof...(T)), error>::type
	encodeMembers(typename Message<schema::Tuple<T...>, Container>::Reader data,
				  Buffer &buffer) {
		error err =
			ProtoKelEncodeImpl<typename Container::template ElementType<i>>::
				encode(data.template get<i>(), buffer);
		if (err.failed()) {
			return err;
		}

		return encodeMembers<i + 1>(data, buffer);
	}

	static error
	encode(typename Message<schema::Tuple<T...>, Container>::Reader data,
		   Buffer &buffer) {
		return encodeMembers<0>(data, buffer);
	}

	template <size_t i = 0>
	static typename std::enable_if<i == sizeof...(T), size_t>::type
	sizeMembers(typename Message<schema::Tuple<T...>, Container>::Reader) {
		return 0;
	}

	template <size_t i = 0>
		static typename std::enable_if <
		i<sizeof...(T), size_t>::type sizeMembers(
			typename Message<schema::Tuple<T...>, Container>::Reader reader) {
		return ProtoKelEncodeImpl<typename Container::template ElementType<i>>::
				   size(reader.template get<i>()) +
			   sizeMembers<i + 1>(reader);
	}

	static size_t
	size(typename Message<schema::Tuple<T...>, Container>::Reader reader) {
		return sizeMembers<0>(reader);
	}
};

template <typename... V, StringLiteral... K, class Container>
struct ProtoKelEncodeImpl<
	Message<schema::Struct<schema::NamedMember<V, K>...>, Container>> {
	template <size_t i = 0>
	static typename std::enable_if<i == sizeof...(V), error>::type
	encodeMembers(typename Message<schema::Struct<schema::NamedMember<V, K>...>,
								   Container>::Reader,
				  Buffer &) {
		return no_error();
	}
	template <size_t i = 0>
		static typename std::enable_if <
		i<sizeof...(V), error>::type encodeMembers(
			typename Message<schema::Struct<schema::NamedMember<V, K>...>,
							 Container>::Reader data,
			Buffer &buffer) {

		error err =
			ProtoKelEncodeImpl<typename Container::template ElementType<i>>::
				encode(data.template get<i>(), buffer);
		if (err.failed()) {
			return err;
		}
		return encodeMembers<i + 1>(data, buffer);
	}

	static error
	encode(typename Message<schema::Struct<schema::NamedMember<V, K>...>,
							Container>::Reader data,
		   Buffer &buffer) {
		return encodeMembers<0>(data, buffer);
	}

	template <size_t i = 0>
	static typename std::enable_if<i == sizeof...(V), size_t>::type
	sizeMembers(typename Message<schema::Struct<schema::NamedMember<V, K>...>,
								 Container>::Reader) {
		return 0;
	}

	template <size_t i = 0>
		static typename std::enable_if <
		i<sizeof...(V), size_t>::type sizeMembers(
			typename Message<schema::Struct<schema::NamedMember<V, K>...>,
							 Container>::Reader reader) {
		return ProtoKelEncodeImpl<typename Container::template ElementType<i>>::
				   size(reader.template get<i>()) +
			   sizeMembers<i + 1>(reader);
	}

	static size_t
	size(typename Message<schema::Struct<schema::NamedMember<V, K>...>,
						  Container>::Reader reader) {
		return sizeMembers<0>(reader);
	}
};

template <typename... V, StringLiteral... K, class Container>
struct ProtoKelEncodeImpl<
	Message<schema::Union<schema::NamedMember<V, K>...>, Container>> {

	template <size_t i = 0>
	static typename std::enable_if<i == sizeof...(V), error>::type
	encodeMembers(typename Message<schema::Union<schema::NamedMember<V, K>...>,
								   Container>::Reader,
				  Buffer &) {
		return no_error();
	}

	template <size_t i = 0>
		static typename std::enable_if <
		i<sizeof...(V), error>::type encodeMembers(
			typename Message<schema::Union<schema::NamedMember<V, K>...>,
							 Container>::Reader reader,
			Buffer &buffer) {
		if (reader.index() == i) {
			error err = StreamValue<ProtoKelCodec::UnionIdT>::encode(i, buffer);
			if (err.failed()) {
				return err;
			}
			return ProtoKelEncodeImpl<typename Container::template ElementType<
				i>>::encode(reader.template get<i>(), buffer);
		}
		return encodeMembers<i + 1>(reader, buffer);
	}

	static error
	encode(typename Message<schema::Union<schema::NamedMember<V, K>...>,
							Container>::Reader reader,
		   Buffer &buffer) {
		return encodeMembers<0>(reader, buffer);
	}

	template <size_t i = 0>
	static typename std::enable_if<i == sizeof...(V), size_t>::type
	sizeMembers(typename Message<schema::Union<schema::NamedMember<V, K>...>,
								 Container>::Reader) {
		return 0;
	}

	template <size_t i = 0>
		static typename std::enable_if <
		i<sizeof...(V), size_t>::type sizeMembers(
			typename Message<schema::Union<schema::NamedMember<V, K>...>,
							 Container>::Reader reader) {
		if (reader.index() == i) {
			return ProtoKelEncodeImpl<typename Container::template ElementType<
				i>>::size(reader.template get<i>());
		}
		return sizeMembers<i + 1>(reader);
	}

	/*
	 * Size of union id + member size
	 */
	static size_t
	size(typename Message<schema::Union<schema::NamedMember<V, K>...>,
						  Container>::Reader reader) {
		return sizeof(ProtoKelCodec::UnionIdT) + sizeMembers<0>(reader);
	}
};

template <class T, class Container>
struct ProtoKelEncodeImpl<Message<schema::Array<T>, Container>> {
	static error
	encode(typename Message<schema::Array<T>, Container>::Reader data,
		   Buffer &buffer) {
		ProtoKelCodec::ArrayLengthT array_length = data.size();
		{
			error err = StreamValue<ProtoKelCodec::ArrayLengthT>::encode(
				array_length, buffer);
			if (err.failed()) {
				return err;
			}
		}

		for (size_t i = 0; i < array_length; ++i) {
			error err =
				ProtoKelEncodeImpl<typename Container::ElementType>::encode(
					data.get(i), buffer);
			if (err.failed()) {
				return err;
			}
		}
		return no_error();
	}

	/*
	 *
	 */
	static size_t
	size(typename Message<schema::Array<T>, Container>::Reader data) {
		size_t members = sizeof(ProtoKelCodec::ArrayLengthT);
		for (size_t i = 0; i < data.size(); ++i) {
			members +=
				ProtoKelEncodeImpl<typename Container::ElementType>::size(
					data.get(i));
		}

		return members;
	}
};

/*
 * Decode Implementations
 */
template <typename T> struct ProtoKelDecodeImpl;

template <class T, size_t N, class Container>
struct ProtoKelDecodeImpl<Message<schema::Primitive<T, N>, Container>> {
	static error
	decode(typename Message<schema::Primitive<T, N>, Container>::Builder data,
		   Buffer &buffer) {
		typename PrimitiveTypeHelper<schema::Primitive<T, N>>::Type val = 0;
		error err = StreamValue<typename PrimitiveTypeHelper<
			schema::Primitive<T, N>>::Type>::decode(val, buffer);
		data.set(val);
		return err;
	}
};

template <class Container>
struct ProtoKelDecodeImpl<Message<schema::String, Container>> {
	static error
	decode(typename Message<schema::String, Container>::Builder data,
		   Buffer &buffer) {
		size_t size = 0;
		if (sizeof(size) > buffer.readCompositeLength()) {
			return recoverable_error("Buffer too small");
		}

		error err = StreamValue<size_t>::decode(size, buffer);
		if (err.failed()) {
			return err;
		}

		if (size > buffer.readCompositeLength()) {
			return recoverable_error("Buffer too small");
		}

		std::string value;
		value.resize(size);

		if (size > buffer.readCompositeLength()) {
			return recoverable_error("Buffer too small");
		}
		for (size_t i = 0; i < value.size(); ++i) {
			value[i] = buffer.read(i);
		}
		buffer.readAdvance(value.size());
		data.set(std::move(value));
		return no_error();
	}
};

template <class... T, class Container>
struct ProtoKelDecodeImpl<Message<schema::Tuple<T...>, Container>> {
	template <size_t i = 0>
	static typename std::enable_if<i == sizeof...(T), error>::type
	decodeMembers(typename Message<schema::Tuple<T...>, Container>::Builder,
				  Buffer &) {
		return no_error();
	}

	template <size_t i = 0>
		static typename std::enable_if <
		i<sizeof...(T), error>::type decodeMembers(
			typename Message<schema::Tuple<T...>, Container>::Builder builder,
			Buffer &buffer) {

		error err =
			ProtoKelDecodeImpl<typename Container::template ElementType<i>>::
				decode(builder.template init<i>(), buffer);
		if (err.failed()) {
			return err;
		}
		return decodeMembers<i + 1>(builder, buffer);
	}

	static error
	decode(typename Message<schema::Tuple<T...>, Container>::Builder builder,
		   Buffer &buffer) {
		return decodeMembers<0>(builder, buffer);
	}
};

template <class... V, StringLiteral... K, class Container>
struct ProtoKelDecodeImpl<
	Message<schema::Struct<schema::NamedMember<V, K>...>, Container>> {

	template <size_t i = 0>
	static typename std::enable_if<i == sizeof...(V), error>::type
	decodeMembers(typename Message<schema::Struct<schema::NamedMember<V, K>...>,
								   Container>::Builder,
				  Buffer &) {
		return no_error();
	}

	template <size_t i = 0>
		static typename std::enable_if <
		i<sizeof...(V), error>::type decodeMembers(
			typename Message<schema::Struct<schema::NamedMember<V, K>...>,
							 Container>::Builder builder,
			Buffer &buffer) {

		error err =
			ProtoKelDecodeImpl<typename Container::template ElementType<i>>::
				decode(builder.template init<i>(), buffer);
		if (err.failed()) {
			return err;
		}
		return decodeMembers<i + 1>(builder, buffer);
	}

	static error
	decode(typename Message<schema::Struct<schema::NamedMember<V, K>...>,
							Container>::Builder builder,
		   Buffer &buffer) {
		return decodeMembers<0>(builder, buffer);
	}
};

template <class... V, StringLiteral... K, class Container>
struct ProtoKelDecodeImpl<
	Message<schema::Union<schema::NamedMember<V, K>...>, Container>> {
	template <size_t i = 0>
	static typename std::enable_if<i == sizeof...(V), error>::type
	decodeMembers(typename Message<schema::Union<schema::NamedMember<V, K>...>,
								   Container>::Builder,
				  Buffer &, ProtoKelCodec::UnionIdT) {
		return no_error();
	}

	template <size_t i = 0>
		static typename std::enable_if <
		i<sizeof...(V), error>::type decodeMembers(
			typename Message<schema::Union<schema::NamedMember<V, K>...>,
							 Container>::Builder builder,
			Buffer &buffer, ProtoKelCodec::UnionIdT id) {

		if (id == i) {
			error err =
				ProtoKelDecodeImpl<typename Container::template ElementType<
					i>>::decode(builder.template init<i>(), buffer);
			if (err.failed()) {
				return err;
			}
		}
		return decodeMembers<i + 1>(builder, buffer, id);
	}

	static error
	decode(typename Message<schema::Union<schema::NamedMember<V, K>...>,
							Container>::Builder builder,
		   Buffer &buffer) {
		ProtoKelCodec::UnionIdT id = 0;
		error err = StreamValue<ProtoKelCodec::UnionIdT>::decode(id, buffer);
		if (err.failed()) {
			return err;
		}
		if (id >= sizeof...(V)) {
			return critical_error("Union doesn't have this many id's");
		}

		return decodeMembers<0>(builder, buffer, id);
	}
};

template <class T, class Container>
struct ProtoKelDecodeImpl<Message<schema::Array<T>, Container>> {
	static error
	decode(typename Message<schema::Array<T>, Container>::Builder data,
		   Buffer &buffer) {
		ProtoKelCodec::ArrayLengthT array_length = 0;
		{
			error err = StreamValue<ProtoKelCodec::ArrayLengthT>::decode(
				array_length, buffer);
			if (err.failed()) {
				return err;
			}
		}

		data.resize(array_length);

		for (size_t i = 0; i < array_length; ++i) {
			error err =
				ProtoKelDecodeImpl<typename Container::ElementType>::decode(
					data.init(i), buffer);
			if (err.failed()) {
				return err;
			}
		}
		return no_error();
	}
};

template <class Schema, class Container>
error ProtoKelCodec::encode(typename Message<Schema, Container>::Reader reader,
							Buffer &buffer) {
	BufferView view{buffer};

	ProtoKelCodec::PacketLengthT packet_length =
		ProtoKelEncodeImpl<Message<Schema, Container>>::size(reader);
	// Check the size of the packet for the first
	// message length description

	error err = view.writeRequireLength(packet_length +
										sizeof(ProtoKelCodec::PacketLengthT));
	if (err.failed()) {
		return err;
	}

	{
		error err = StreamValue<ProtoKelCodec::PacketLengthT>::encode(
			packet_length, view);
		if (err.failed()) {
			return err;
		}
	}
	{
		error err = ProtoKelEncodeImpl<Message<Schema, Container>>::encode(
			reader, view);
		if (err.failed()) {
			return err;
		}
	}

	buffer.writeAdvance(view.writeOffset());
	return no_error();
}

template <class Schema, class Container>
error ProtoKelCodec::decode(
	typename Message<Schema, Container>::Builder builder, Buffer &buffer,
	const Limits &limits) {
	BufferView view{buffer};

	ProtoKelCodec::PacketLengthT packet_length = 0;
	{
		error err = StreamValue<ProtoKelCodec::PacketLengthT>::decode(
			packet_length, view);
		if (err.failed()) {
			return err;
		}
	}

	if (packet_length > limits.packet_size) {
		return critical_error(
			[packet_length]() {
				return std::string{"Packet size too big: "} +
					   std::to_string(packet_length);
			},
			"Packet size too big");
	}

	{
		error err = ProtoKelDecodeImpl<Message<Schema, Container>>::decode(
			builder, view);
		if (err.failed()) {
			return err;
		}
	}
	{
		if (ProtoKelEncodeImpl<Message<Schema, Container>>::size(
				builder.asReader()) != packet_length) {
			return critical_error("Bad packet format");
		}
	}

	buffer.readAdvance(view.readOffset());
	return no_error();
}

} // namespace saw
