#include "tls.h"

#include <gnutls/gnutls.h>
#include <gnutls/x509.h>

#include "io_helpers.h"

#include <cassert>

#include <iostream>

namespace saw {

class Tls::Impl {
public:
	gnutls_certificate_credentials_t xcred;

public:
	Impl() {
		gnutls_global_init();
		gnutls_certificate_allocate_credentials(&xcred);
		gnutls_certificate_set_x509_system_trust(xcred);
	}

	~Impl() {
		gnutls_certificate_free_credentials(xcred);
		gnutls_global_deinit();
	}
};

static ssize_t forst_tls_push_func(gnutls_transport_ptr_t p, const void *data,
						 size_t size);
static ssize_t forst_tls_pull_func(gnutls_transport_ptr_t p, void *data, size_t size);

Tls::Tls() : impl{heap<Tls::Impl>()} {}

Tls::~Tls() {}

Tls::Impl &Tls::getImpl() { return *impl; }

class TlsIoStream final : public IoStream {
private:
	own<IoStream> internal;
	gnutls_session_t session_handle;

public:
	TlsIoStream(own<IoStream> internal_) : internal{std::move(internal_)} {}

	~TlsIoStream() { gnutls_bye(session_handle, GNUTLS_SHUT_RDWR); }

	error_or<size_t> read(void *buffer, size_t length) override {
		ssize_t size = gnutls_record_recv(session_handle, buffer, length);
		if (size < 0) {
			if(gnutls_error_is_fatal(size) == 0){
				return recoverable_error([size](){return std::string{"Read recoverable error "}+std::string{gnutls_strerror(size)};}, "error read r");
			}else{
				return critical_error([size](){return std::string{"Read critical error "}+std::string{gnutls_strerror(size)};}, "error read c");
			}
		}else if(size == 0){
			return critical_error("Disconnected");
		}

		return static_cast<size_t>(length);
	}

	conveyor<void> readReady() override { return internal->readReady(); }

	conveyor<void> onReadDisconnected() override {
		return internal->onReadDisconnected();
	}

	error_or<size_t> write(const void *buffer, size_t length) override {
		ssize_t size = gnutls_record_send(session_handle, buffer, length);
		if(size < 0){
			if(gnutls_error_is_fatal(size) == 0){
				return recoverable_error([size](){return std::string{"Write recoverable error "}+std::string{gnutls_strerror(size)} + " " + std::to_string(size);}, "error write r");
			}else{
				return critical_error([size](){return std::string{"Write critical error "}+std::string{gnutls_strerror(size)} + " " + std::to_string(size);}, "error write c");
			}
		}

		return static_cast<size_t>(size);
	}

	conveyor<void> writeReady() override { return internal->writeReady(); }

	gnutls_session_t &session() { return session_handle; }
};

TlsServer::TlsServer(own<Server> srv) : internal{std::move(srv)} {}

conveyor<own<IoStream>> TlsServer::accept() {
	SAW_ASSERT(internal) { return conveyor<own<IoStream>>{fix_void<own<IoStream>>{nullptr}}; }
	return internal->accept().then([](own<IoStream> stream) -> own<IoStream> {
		/// @todo handshake


		return heap<TlsIoStream>(std::move(stream));
	});
}

namespace {
/*
* Small helper for setting up the nonblocking connection handshake
*/
struct TlsClientStreamHelper {
public:
	own<conveyor_feeder<own<IoStream>>> feeder;
	conveyor_sink connection_sink;
	conveyor_sink stream_reader;
	conveyor_sink stream_writer;

	own<TlsIoStream> stream = nullptr;
public:
	TlsClientStreamHelper(own<conveyor_feeder<own<IoStream>>> f):
		feeder{std::move(f)}
	{}

	void setupTurn(){
		SAW_ASSERT(stream){
			return;
		}

		stream_reader = stream->readReady().then([this](){
			turn();
		}).sink();

		stream_writer = stream->writeReady().then([this](){
			turn();
		}).sink();
	}

	void turn(){
		if(stream){
			// Guarantee that the receiving end is already setup
			SAW_ASSERT(feeder){
				return;
			}

			auto &session = stream->session();

			int ret;
			do {
				ret = gnutls_handshake(session);
			} while ( (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED) && gnutls_error_is_fatal(ret) == 0);

			if(gnutls_error_is_fatal(ret)){
				feeder->fail(critical_error("Couldn't create Tls connection"));
				stream = nullptr;
			}else if(ret == GNUTLS_E_SUCCESS){
				feeder->feed(std::move(stream));
			}
		}
	}
};
}

own<Server> TlsNetwork::listen(NetworkAddress& address) {
	return heap<TlsServer>(internal.listen(address));
}

conveyor<own<IoStream>> TlsNetwork::connect(NetworkAddress& address) {
	// Helper setups
	auto caf = new_conveyor_and_feeder<own<IoStream>>();
	own<TlsClientStreamHelper> helper = heap<TlsClientStreamHelper>(std::move(caf.feeder));
	TlsClientStreamHelper* hlp_ptr = helper.get();
	
	// conveyor entangled structure
	auto prim_conv = internal.connect(address).then([this, hlp_ptr, addr = address.address()](
										own<IoStream> stream) -> error_or<void> {
		IoStream* inner_stream = stream.get();
		auto tls_stream = heap<TlsIoStream>(std::move(stream));

		auto &session = tls_stream->session();

		gnutls_init(&session, GNUTLS_CLIENT);

		gnutls_server_name_set(session, GNUTLS_NAME_DNS, addr.c_str(),
							   addr.size());

		gnutls_set_default_priority(session);
		gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE,
							   tls.getImpl().xcred);
		gnutls_session_set_verify_cert(session, addr.c_str(), 0);

		gnutls_transport_set_ptr(session, reinterpret_cast<gnutls_transport_ptr_t>(inner_stream));
		gnutls_transport_set_push_function(session, forst_tls_push_func);
		gnutls_transport_set_pull_function(session, forst_tls_pull_func);

		// gnutls_handshake_set_timeout(session, GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT);

		hlp_ptr->stream = std::move(tls_stream);
		hlp_ptr->setupTurn();
		hlp_ptr->turn();

		return Void{};
	});

	helper->connection_sink = prim_conv.sink();

	return caf.conveyor.attach(std::move(helper));
}

own<Datagram> TlsNetwork::datagram(NetworkAddress& address){
	///@unimplemented
	return nullptr;
}

static ssize_t forst_tls_push_func(gnutls_transport_ptr_t p, const void *data,
						 size_t size) {
	IoStream *stream = reinterpret_cast<IoStream *>(p);
	if (!stream) {
		return -1;
	}

	error_or<size_t> length = stream->write(data, size);
	if (length.is_error() || !length.is_value()) {
		return -1;
	}

	return static_cast<ssize_t>(length.value());
}

static ssize_t forst_tls_pull_func(gnutls_transport_ptr_t p, void *data, size_t size) {
	IoStream *stream = reinterpret_cast<IoStream *>(p);
	if (!stream) {
		return -1;
	}

	error_or<size_t> length = stream->read(data, size);
	if (length.is_error() || !length.is_value()) {
		return -1;
	}

	return static_cast<ssize_t>(length.value());
}

TlsNetwork::TlsNetwork(Tls& tls_, Network &network) : tls{tls_},internal{network} {}

conveyor<own<NetworkAddress>> TlsNetwork::resolveAddress(const std::string &addr,
													   uint16_t port) {
	/// @todo tls server name needed. Check validity. Won't matter later on, because gnutls should fail anyway. But
	/// it's better to find the error source sooner rather than later
	return internal.resolveAddress(addr, port);
}

std::optional<own<TlsNetwork>> setupTlsNetwork(Network &network) {
	return std::nullopt;
}
} // namespace saw
